//Copyright (c) 2010 MPL1031
//
//Permission is hereby granted, free of charge, to any person obtaining a copy
//of this software and associated documentation files (the "Software"), to de$
//in the Software without restriction, including without limitation the rights
//to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//copies of the Software, and to permit persons to whom the Software is
//furnished to do so, subject to the following conditions:
//
//The above copyright notice and this permission notice shall be included in
//all copies or substantial portions of the Software.
//
//THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FRO$
//OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//THE SOFTWARE.


package org.chargecar.prize.policies;

import java.io.*;
import java.util.*;
import org.chargecar.prize.battery.BatteryModel;
import org.chargecar.prize.util.PointFeatures;
import org.chargecar.prize.util.PowerFlowException;
import org.chargecar.prize.util.PowerFlows;
import org.chargecar.prize.util.TripFeatures;

/**
 * This policy generates battery target watt values for all data points on a trip.
 * If the location is known, the pointTargetMap is used.  If the location isn't known,
 * then battPowerTargets are used.
 *
 * 2010-12 training data: 54.893% reduction
 * 2010-12 judging data: 51.911% reduction
 * 
 * @author MPL1031@gmail.com
 */
public class MPL1031PolicyV6 implements Policy {

	private final static String name = "MPL1031 Policy v6";
	private BatteryModel modelCap;
	private BatteryModel modelBatt;
	private double capTotalWattsCapacity;

	private static Map<String,Double> alikPointTargetMap;
	private static Map<String,Double> arnoldPointTargetMap;
	private static Map<String,Double> garyPointTargetMap;
	private static Map<String,Double> illahPointTargetMap;
	private static Map<String,Double> mikePointTargetMap;
	private static Map<String,Double> nikolayPointTargetMap;
	private static Map<String,Double> richPointTargetMap;
	private static Map<String,Double> thorPointTargetMap;
	private static Map<String,Double> pointTargetMap;
	private double priorLat;
	private double priorLon;
	private double priorLatitude;
	private double priorLongitude;
	private static final double UNKNOWN = Double.MIN_VALUE;
	private String compassPoint;
	private double targetWatts;

	private double[][] alikBattPowerTargets = new double[][] {
		{-10000,-10000,-10000,-10000,-15243,-6982,-6873,-8847,-11484,-4428,-4467,-4468,-6022,-13756,-15339,-9004,-10078,-13873,-9856,-8728,-9818,},  // speed: 0
		{-10000,-10000,-10000,-20430,-9341,-11562,-13982,-12724,-14158,-15391,-9199,-7514,-9350,-15003,-15400,-17547,-15184,-9892,-30895,-17479,-12583,},  // speed: 1
		{-10000,-10000,-8385,-9958,-11301,-13995,-11082,-14189,-13076,-17694,-15119,-12171,-14238,-25269,-19716,-20053,-45845,-19172,-25323,-15210,-13279,},  // speed: 2
		{-15000,-15000,-10727,-9533,-14932,-13826,-11860,-13906,-14840,-14131,-19853,-15867,-9448,-12179,-14167,-17706,-23117,-40403,-69570,-13172,-12782,},  // speed: 3
		{-15000,-8146,-18667,-17061,-15027,-15378,-10669,-12763,-14419,-12391,-14558,-12531,-16230,-15298,-18712,-17501,-19947,-42744,-46855,-19436,-11871,},  // speed: 4
		{-15000,-6971,-3190,-5935,-13058,-8864,-11100,-13164,-9806,-11030,-13625,-13373,-11484,-14892,-14506,-14814,-19357,-28524,-36624,-20863,-13485,},  // speed: 5
		{-15000,-8517,-16734,-9335,-12217,-9663,-9771,-8612,-9147,-13308,-12054,-8157,-9549,-12627,-12995,-12905,-18146,-25132,-41344,-35449,-11217,},  // speed: 6
		{-19740,-7517,-12949,-11987,-13570,-8257,-15710,-10519,-10450,-16021,-15212,-12540,-10556,-17206,-13657,-13567,-20908,-18536,-31979,-18024,-33645,},  // speed: 7
		{-10613,-33903,-9544,-9147,-12585,-15836,-10126,-12441,-12793,-11015,-10762,-13309,-11852,-16711,-12570,-15591,-15901,-16939,-37500,-35185,-30924,},  // speed: 8
		{-6575,-19460,-20958,-9601,-18031,-10846,-13135,-11398,-9456,-11479,-11760,-9370,-9468,-13777,-13008,-12209,-16810,-18004,-20731,-71747,-46448,},  // speed: 9
		{-11154,-10938,-22941,-7740,-10802,-11000,-11260,-8815,-9341,-15129,-9343,-12008,-14581,-14809,-13338,-12954,-14918,-20053,-24882,-35207,-38109,},  // speed: 10
		{-49458,-25980,-9501,-8590,-9223,-8138,-7984,-9213,-14242,-9335,-9159,-12362,-14183,-12617,-10570,-12252,-14329,-18964,-24451,-24233,-42586,},  // speed: 11
		{-27927,-12676,-12751,-9842,-11340,-7611,-8552,-10324,-8976,-12212,-9571,-8464,-10164,-8984,-11253,-11323,-13336,-15449,-17764,-26484,-42068,},  // speed: 12
		{-23167,-15346,-9292,-14020,-14226,-10367,-9685,-9113,-10245,-8792,-7198,-11480,-10453,-8137,-11443,-11358,-12276,-14448,-22521,-25338,-23816,},  // speed: 13
		{-15822,-14128,-11749,-13436,-14059,-13544,-11793,-9568,-9525,-11108,-28465,-10304,-11849,-14190,-15575,-13131,-14248,-18806,-17517,-25711,-33524,},  // speed: 14
		{-7064,-12590,-24493,-11458,-9614,-10960,-10127,-14888,-11103,-12488,-8178,-10781,-16452,-12631,-13298,-12749,-14934,-16843,-16580,-18573,-17757,},  // speed: 15
		{-5154,-16692,-18578,-11872,-12884,-11934,-16316,-25033,-9951,-16863,-8955,-10158,-11897,-9226,-11534,-14030,-16656,-19280,-25297,-21726,-24961,},  // speed: 16
		{-36691,-19442,-6002,-12637,-15701,-13016,-20191,-18144,-19449,-13125,-10601,-16355,-12308,-15028,-16834,-14368,-16599,-21968,-17144,-13535,-13966,},  // speed: 17
		{-9163,-17624,0,-7762,-18973,-16251,-21602,-13226,-12080,-20822,-3135,-12181,-24425,-21974,-25941,-14307,-17040,-20653,-24772,-60222,-53238,},  // speed: 18
		{-14379,-35546,-12791,-14029,-16149,-20601,-16265,-18082,-21111,-13506,-20718,-11145,-17922,-20178,-18472,-17696,-20848,-21890,-25607,-35873,-15959,},  // speed: 19
		{-9145,-4662,-25697,-12708,-17355,-15809,-18725,-17417,-16477,-17869,-12503,-14473,-14835,-19066,-21160,-19657,-21855,-23832,-26795,-34758,-5312,},  // speed: 20
		{-7781,-3290,-12818,-17843,-15969,-18133,-15665,-12812,-20028,-21474,-15448,-28483,-16853,-24895,-19303,-21798,-22230,-26805,-28584,-35741,-44793,},  // speed: 21
		{-28346,-4487,-22767,-10993,-21553,-16525,-14095,-17530,-16947,-19733,-19723,-24824,-15464,-21692,-24068,-23546,-20335,-26741,-28667,-32003,-29008,},  // speed: 22
		{-10000,-13251,-19374,-20478,-21647,-16416,-16563,-17168,-18931,-17197,-17125,-15523,-21486,-22728,-19605,-19585,-22084,-24175,-29271,-29597,-25322,},  // speed: 23
		{-17583,-26702,-9738,-19892,-20441,-17677,-18009,-21841,-22382,-21034,-14527,-37012,-22883,-18871,-22456,-23153,-24715,-25935,-32584,-32279,-18369,},  // speed: 24
		{-25919,-22098,-26640,-22652,-21260,-21682,-22114,-20172,-21673,-20814,-20545,-26790,-19818,-21213,-22860,-22919,-23733,-28067,-31238,-38003,-25441,},  // speed: 25
		{-43931,-23200,-27619,-24232,-20926,-22972,-20854,-22011,-21272,-23641,-22317,-24853,-19091,-25173,-21019,-24093,-25283,-28315,-29561,-34583,-25417,},  // speed: 26
		{-19142,-24431,-22107,-28224,-24030,-22908,-28444,-25614,-17820,-23662,-31611,-26899,-23180,-22878,-28558,-24882,-26672,-27868,-30817,-39032,-58360,},  // speed: 27
		{-25000,-19682,-27101,-21800,-22127,-23237,-22425,-21919,-24117,-26664,-28798,-27119,-22703,-27587,-26593,-27786,-27368,-29292,-33611,-35634,-47266,},  // speed: 28
		{-25836,-23903,-27960,-31126,-27075,-20853,-23472,-23278,-27897,-20495,-20712,-27301,-26826,-22580,-25023,-25052,-26541,-27736,-29790,-36523,-37162,},  // speed: 29
		{-58193,-42978,-25221,-20987,-26260,-36357,-26479,-25004,-18173,-22177,-21812,-26447,-19393,-22633,-25447,-25277,-25665,-28448,-31526,-37918,-46713,},  // speed: 30
		{-16539,-25454,-17293,-29769,-28937,-24638,-19458,-15464,-17423,-21179,-21507,-42405,-24301,-23104,-24844,-23353,-25593,-28250,-30564,-37328,-44686,},  // speed: 31
		{-25000,-25000,-25000,-35291,-24271,-27361,-26194,-26902,-22838,-22050,-25633,-19467,-29186,-21390,-19080,-26337,-27161,-27343,-32331,-30356,-46249,},  // speed: 32
		{-25000,-25000,-25000,-25000,-17411,-14788,-13709,-22938,-30802,-30654,-20172,-20840,-49770,-29826,-41029,-25073,-28539,-31592,-37674,-36819,-36536,},  // speed: 33
		{-25000,-25000,-18993,-17039,-17609,-18277,-25450,-14949,-15000,-15000,-15000,-18561,-16766,-25278,-25986,-19885,-23436,-29290,-30608,-36756,-20000,},  // speed: 34
	};
	private double[][] arnoldBattPowerTargets = new double[][] {
		{-10000,-10000,-10000,-10000,-6147,-7967,-7843,-10432,-11337,-8190,-7811,-8264,-10000,-8921,-11428,-10060,-11666,-13873,-9856,-8728,-9818,},  // speed: 0
		{-10000,-10000,-10000,-28603,-8172,-10239,-10073,-9563,-9781,-11904,-11891,-10047,-11425,-11314,-13227,-13073,-12232,-13347,-28119,-17479,-12583,},  // speed: 1
		{-10000,-10000,-8385,-10445,-8892,-10280,-11041,-10540,-10979,-10709,-16850,-11355,-13717,-13161,-15651,-14464,-16712,-16835,-12132,-15210,-13279,},  // speed: 2
		{-15000,-15000,-6903,-10415,-10784,-10413,-9858,-10869,-10489,-11507,-10383,-10994,-11695,-13219,-14014,-17023,-17990,-18107,-7791,-13172,-12782,},  // speed: 3
		{-15000,-8146,-6918,-11338,-9782,-9507,-10254,-12293,-9819,-10741,-10528,-15480,-10630,-11237,-13333,-15461,-18558,-18531,-14965,-19436,-11871,},  // speed: 4
		{-15000,-6688,-13847,-10770,-9684,-9225,-9523,-11455,-9725,-13544,-9775,-10308,-10704,-12377,-15145,-14034,-18091,-19957,-19944,-9046,-13485,},  // speed: 5
		{-15000,-8517,-31188,-9838,-9179,-8977,-9896,-9954,-10037,-12093,-11004,-10107,-13147,-11317,-12452,-13277,-14650,-17809,-21278,-17573,-11217,},  // speed: 6
		{-20489,-11823,-8863,-9876,-9067,-8926,-9164,-9660,-9101,-10808,-10156,-9984,-9884,-10487,-11534,-11788,-13406,-16417,-21919,-19339,-16376,},  // speed: 7
		{-10613,-26765,-24516,-9036,-8700,-8468,-9110,-8713,-9369,-9850,-8857,-9047,-9500,-10380,-10663,-11725,-12381,-15618,-18555,-26162,-17266,},  // speed: 8
		{-22680,-22832,-6579,-9160,-8014,-8215,-8884,-8433,-8984,-8765,-10590,-8414,-8792,-9634,-9450,-10421,-11943,-13387,-17121,-30968,-51991,},  // speed: 9
		{-20705,-30529,-11259,-8375,-8461,-8041,-8520,-8313,-9165,-9025,-9238,-8509,-9172,-9055,-8852,-9741,-11038,-13048,-17296,-24793,-18459,},  // speed: 10
		{-52446,-51026,-15898,-8756,-8259,-8251,-8169,-8170,-8032,-8362,-8144,-7960,-8198,-8335,-8527,-8837,-10293,-12344,-17633,-19967,-37127,},  // speed: 11
		{-27927,-15952,-13743,-7499,-8018,-8646,-7641,-7740,-7914,-7486,-8040,-8617,-8168,-7988,-8278,-8356,-9517,-11323,-15981,-31102,-27382,},  // speed: 12
		{-4860,-20644,-9419,-7898,-8148,-7974,-7815,-8072,-7693,-7900,-7310,-7633,-8075,-7896,-8256,-8813,-9657,-11835,-16751,-20672,-23606,},  // speed: 13
		{-6034,-5832,-7817,-8084,-6842,-7449,-7675,-7785,-8034,-7899,-7605,-7957,-8053,-7846,-8274,-8519,-9882,-11936,-17639,-30351,-32746,},  // speed: 14
		{-4787,-17072,-5967,-7649,-7422,-7522,-7812,-7823,-7337,-7801,-7021,-8105,-7800,-8030,-8656,-8689,-9962,-13063,-18433,-29135,-23929,},  // speed: 15
		{-22925,-8323,-3463,-8094,-9378,-9330,-8599,-8456,-8457,-8449,-7977,-7929,-8501,-8702,-9805,-9956,-11592,-14361,-18331,-20679,-41609,},  // speed: 16
		{-57289,-9659,-5546,-5886,-8393,-10050,-9489,-8834,-8944,-8814,-9888,-8184,-8697,-9298,-9626,-10866,-11932,-14010,-18574,-20098,-42586,},  // speed: 17
		{-6373,-11913,-3779,-6612,-10451,-9303,-10026,-9561,-10412,-10201,-10280,-10353,-9622,-11019,-10547,-11984,-13086,-15730,-20838,-29789,-16659,},  // speed: 18
		{-5395,-34483,-3826,-7813,-9107,-9592,-9309,-10301,-11369,-12352,-10496,-10310,-10816,-11190,-13065,-12699,-15328,-16859,-21334,-28098,-43043,},  // speed: 19
		{-3037,-7538,-6949,-5681,-8185,-10143,-9989,-11372,-11032,-11340,-13238,-13569,-13242,-12659,-13393,-14351,-16567,-18704,-22103,-26033,-5312,},  // speed: 20
		{-6558,0,-3673,-5991,-10124,-11549,-11366,-12765,-12941,-13524,-12519,-14482,-14452,-14337,-16197,-15441,-17326,-19787,-21902,-24788,-39628,},  // speed: 21
		{-45500,-396,-10036,-9290,-9530,-10523,-12929,-12221,-13579,-13684,-12584,-14324,-13314,-13691,-15135,-16542,-17491,-19565,-22190,-20033,0,},  // speed: 22
		{-10000,0,-6160,-11850,-11378,-13044,-13471,-13707,-12903,-14039,-15624,-13457,-15166,-14987,-16605,-16882,-18736,-20276,-22209,-23825,-21043,},  // speed: 23
		{-32288,0,-8575,-10800,-11800,-12351,-12974,-13229,-15288,-14636,-15074,-15158,-15153,-15243,-15875,-17064,-18462,-20791,-22517,-33005,-18369,},  // speed: 24
		{-8749,-43471,-1582,-12633,-12024,-15227,-14690,-15395,-15939,-15747,-15555,-13929,-15448,-16306,-16143,-17838,-19336,-21430,-23406,-27618,-15943,},  // speed: 25
		{-36636,-10343,-12389,-14456,-15169,-15919,-15319,-18490,-16352,-15930,-15465,-16960,-16885,-17154,-18148,-19472,-20755,-22417,-23991,-26536,-31247,},  // speed: 26
		{-49054,-45922,-23896,-12048,-16854,-16668,-16391,-17074,-17191,-17011,-18177,-17821,-17612,-18308,-18987,-19725,-21102,-22898,-24318,-26455,-42674,},  // speed: 27
		{-29071,-15271,-14272,-19772,-18449,-16632,-17226,-18272,-18409,-17624,-18475,-18170,-19015,-18711,-19200,-19883,-21269,-22486,-24195,-25880,-26188,},  // speed: 28
		{-18618,-29964,-44458,-21080,-19605,-18134,-17865,-18214,-17443,-20005,-17695,-19609,-18939,-19315,-20360,-20278,-21449,-22872,-24123,-26326,-47605,},  // speed: 29
		{-27412,-42978,-19684,-20792,-18965,-19627,-17635,-20721,-23315,-19298,-19016,-20196,-23828,-20652,-19947,-21176,-22226,-23063,-24644,-30503,-22073,},  // speed: 30
		{-14875,-17501,-26554,-15702,-23934,-18712,-18559,-20313,-20358,-19016,-18171,-21755,-18689,-18753,-20902,-20400,-21713,-23813,-25592,-30185,-30628,},  // speed: 31
		{-25000,-25000,-25000,-19708,-17872,-20107,-19585,-18158,-19291,-20692,-19450,-18861,-18753,-16510,-21044,-21745,-21974,-23463,-25863,-26518,-32274,},  // speed: 32
		{-25000,-25000,-25000,-25000,-19639,-18381,-12618,-17396,-14089,-15508,-19753,-22528,-22201,-20165,-24664,-19689,-25068,-24673,-24860,-22935,-26102,},  // speed: 33
		{-25000,-25000,-18375,-15884,-17609,-17524,-22026,-13841,-15000,-15000,-15000,-18561,-16245,-24681,-20000,-20000,-22754,-24834,-28327,-24878,-20000,},  // speed: 34
	};
	private double[][] garyBattPowerTargets = new double[][] {
		{-10000,-10000,-10000,-10000,-12963,-13325,-14399,-14879,-16634,-11651,-13377,-11589,-20258,-17776,-21068,-22166,-25663,-13873,-9856,-8728,-9818,},  // speed: 0
		{-10000,-10000,-10000,-14548,-15357,-15282,-19668,-17781,-13220,-16904,-12070,-19367,-20192,-20181,-19049,-21192,-24197,-27586,-55332,-17479,-12583,},  // speed: 1
		{-10000,-10000,-10039,-10426,-23941,-14973,-17259,-17386,-18952,-17361,-18191,-19595,-17098,-20263,-26253,-32057,-35067,-23018,-131204,-15210,-13279,},  // speed: 2
		{-15000,-15000,-22748,-15251,-15766,-16903,-14539,-16196,-11908,-15167,-19729,-14951,-18154,-26618,-21241,-31779,-30144,-58888,-105999,-13172,-12782,},  // speed: 3
		{-15000,-8146,-9599,-20420,-17538,-17703,-16282,-15668,-23265,-20874,-25261,-9237,-17543,-28469,-24524,-32605,-37089,-50390,-63898,-19436,-11871,},  // speed: 4
		{-15000,-11134,-16170,-14572,-18198,-13753,-16023,-20971,-8732,-16612,-16286,-15258,-21154,-19620,-13004,-27104,-29415,-36959,-83732,-54688,-13485,},  // speed: 5
		{-15000,-30014,-12114,-13758,-17483,-18516,-15323,-15370,-14165,-16187,-28146,-16157,-13382,-29381,-22992,-32218,-32144,-39016,-52945,-67823,-11217,},  // speed: 6
		{-20489,-14926,-19861,-24398,-13859,-14944,-15221,-14435,-7715,-18501,-16051,-24241,-16494,-22983,-21290,-21921,-25209,-33554,-61526,-66059,-16376,},  // speed: 7
		{-17283,-17425,-15986,-15287,-17903,-20763,-14405,-15933,-23563,-19604,-14536,-13335,-19298,-15909,-21906,-18960,-22249,-27804,-44263,-76244,-31080,},  // speed: 8
		{-18039,-23217,-14933,-14773,-15428,-14207,-16313,-18790,-17981,-14395,-14148,-18421,-16294,-15115,-16707,-19865,-17908,-27153,-44956,-53661,-1157,},  // speed: 9
		{-15155,-26246,-14407,-13961,-13209,-12176,-14905,-12530,-12002,-15296,-8998,-14655,-12721,-19593,-15592,-16345,-26159,-24553,-33978,-54019,-48097,},  // speed: 10
		{-9552,-15952,-15211,-18350,-13586,-12436,-13087,-13040,-15634,-11099,-11767,-10374,-11315,-13027,-13041,-14206,-19129,-20555,-27835,-39893,-23198,},  // speed: 11
		{-31931,-22567,-18818,-14129,-13649,-14501,-11780,-9576,-12635,-12880,-12781,-10799,-11610,-13759,-13279,-13623,-16682,-17730,-23767,-45372,-33508,},  // speed: 12
		{-25859,-21090,-12922,-13981,-18273,-12163,-11789,-11427,-12263,-13506,-13984,-10953,-10982,-12866,-14950,-12943,-13570,-18919,-25000,-17158,-23816,},  // speed: 13
		{-15822,-17742,-15798,-13625,-12498,-13073,-11800,-12370,-14305,-14221,-12971,-14230,-13245,-12764,-12730,-13251,-15012,-15486,-20396,-20936,-19959,},  // speed: 14
		{-20941,-19201,-13706,-12957,-11361,-11564,-11194,-10938,-11472,-11552,-14039,-9905,-10759,-11422,-12688,-13717,-16975,-16548,-15779,-18390,-28181,},  // speed: 15
		{-6617,-6483,-8706,-9416,-12249,-11234,-11385,-12475,-11434,-12047,-12257,-11505,-10945,-10697,-13712,-13425,-13969,-15194,-16198,-15447,-30193,},  // speed: 16
		{-36691,-9615,-9545,-11024,-11414,-11488,-10776,-12588,-11552,-12249,-8756,-10582,-11634,-11874,-11986,-14403,-14503,-15579,-16235,-22978,-15841,},  // speed: 17
		{-9163,-7894,-7919,-11300,-12603,-12025,-10389,-11751,-11564,-10229,-10386,-11708,-14172,-16064,-12691,-14209,-15587,-15499,-16382,-17971,-21483,},  // speed: 18
		{-8353,-23891,-8266,-11567,-8797,-11276,-12027,-14323,-18549,-10984,-13193,-14521,-13792,-13401,-13265,-14644,-14740,-17055,-16253,-13167,-15959,},  // speed: 19
		{-9145,-40817,-9979,-7998,-14183,-11256,-13071,-14961,-11640,-13547,-8852,-11907,-9347,-10347,-11377,-13879,-15656,-18434,-19277,-24126,-5312,},  // speed: 20
		{-7781,-4120,-17708,-8823,-10482,-9858,-11686,-12773,-13667,-13781,-16063,-10580,-17175,-12519,-11891,-13670,-17853,-18731,-21071,-23633,-17027,},  // speed: 21
		{-28346,-15467,-17029,-16617,-12723,-10828,-13884,-15552,-13177,-13043,-11178,-17353,-17333,-20466,-16100,-16207,-18991,-20486,-20556,-17111,0,},  // speed: 22
		{-10000,-13251,-9884,-16691,-19150,-16424,-16747,-13441,-15627,-18425,-29655,-16070,-16178,-17402,-17543,-20112,-19777,-23018,-24303,-45141,-25839,},  // speed: 23
		{-17583,-22595,-19263,-15233,-15436,-16277,-21973,-21430,-17524,-19508,-14010,-17232,-20851,-18720,-25454,-20921,-21196,-23622,-25492,-34406,-13401,},  // speed: 24
		{-8749,-22098,-17873,-16205,-19491,-16444,-17873,-21360,-21376,-19872,-23391,-22694,-16251,-20578,-25002,-23589,-24114,-24123,-27081,-28594,-39917,},  // speed: 25
		{-34255,0,-11659,-19271,-17429,-21513,-22387,-19322,-34901,-20266,-21822,-18649,-16254,-27435,-24947,-23065,-26236,-28426,-35734,-44199,-29065,},  // speed: 26
		{-19142,-23562,-12416,-15434,-18565,-23257,-26204,-24278,-16713,-22224,-14341,-28305,-25575,-30711,-28738,-31894,-28673,-32120,-37614,-39526,-23563,},  // speed: 27
		{-25000,-24075,-8036,-15702,-14764,-17046,-20670,-23968,-21858,-22621,-20821,-24487,-18010,-21484,-25478,-30021,-26543,-27807,-30898,-38439,-36705,},  // speed: 28
		{-28703,-29964,-22253,-19039,-19352,-19863,-20182,-18632,-21327,-22466,-26171,-25560,-19013,-21692,-24206,-24346,-26256,-26768,-29487,-36264,-34009,},  // speed: 29
		{-27412,-42978,-27817,-18039,-23173,-19995,-22043,-20998,-24354,-22542,-25886,-20271,-25202,-23681,-22753,-25073,-25507,-27773,-31698,-33901,-28010,},  // speed: 30
		{-14875,-28165,-20487,-23666,-20073,-21320,-13847,-24819,-22036,-18874,-20165,-26674,-24205,-22424,-25688,-25536,-24517,-29128,-32316,-32476,-28466,},  // speed: 31
		{-25000,-33737,-21762,-17531,-22880,-23652,-33092,-19430,-14905,-21983,-25885,-17280,-25465,-21838,-26409,-24164,-27189,-27769,-30964,-33924,-31909,},  // speed: 32
		{-25000,-25000,-14154,-15182,-16396,-20129,-18304,-19622,-19315,-16270,-20172,-22206,-27958,-25168,-23237,-22506,-24831,-26839,-31465,-29576,-44141,},  // speed: 33
		{-25000,-13839,-13050,-16712,-16277,-13119,-13860,-17695,-14565,-15442,-27442,-25824,-10528,-20478,-17177,-19291,-22900,-20594,-24174,-19315,-16675,},  // speed: 34
	};
	private double[][] illahBattPowerTargets = new double[][] {
		{-10000,-10000,-10000,-10000,-17880,-8070,-8520,-8852,-8861,-8473,-8832,-8626,-9451,-10178,-10717,-9896,-10867,-13873,-9856,-8728,-9818,},  // speed: 0
		{-10000,-10000,-10000,-7725,-8945,-8834,-12479,-11129,-10586,-9532,-9032,-9892,-9250,-11989,-10129,-11631,-10671,-15630,-16769,-17479,-12583,},  // speed: 1
		{-10000,-10000,-7582,-6354,-8901,-8547,-9777,-8939,-10744,-9382,-9587,-8441,-9794,-11654,-11661,-13140,-12550,-14800,-18417,-15210,-13279,},  // speed: 2
		{-15000,-15000,-16650,-7940,-9175,-7730,-10386,-11029,-10185,-8763,-9509,-11769,-9750,-11811,-11785,-12430,-13527,-17621,-54074,-13172,-12782,},  // speed: 3
		{-15000,-8146,-7295,-10451,-8887,-8209,-9137,-8030,-7628,-9379,-13697,-9420,-10621,-9607,-10311,-12089,-13790,-17780,-9198,-19436,-11871,},  // speed: 4
		{-15000,-6971,-9284,-8969,-8361,-7804,-8674,-8988,-7891,-9350,-11016,-9818,-8285,-9687,-10010,-11076,-13278,-17070,-21991,-9217,-13485,},  // speed: 5
		{-15000,-6631,-6582,-8952,-7643,-8879,-8196,-7360,-8254,-7533,-8912,-8068,-10617,-7763,-10440,-9441,-11100,-13755,-18179,-28722,-11217,},  // speed: 6
		{-20489,-5875,-13189,-6470,-7514,-7053,-8077,-6793,-7042,-7852,-7556,-7159,-9392,-8331,-8088,-8984,-10457,-14631,-20300,-11740,-16376,},  // speed: 7
		{-10613,-38442,-11181,-9169,-9228,-6291,-7271,-6812,-6827,-6977,-7251,-7152,-7145,-7667,-8392,-9771,-10606,-12889,-18519,-25363,-8771,},  // speed: 8
		{-22680,-4156,-13507,-7681,-6572,-6054,-6064,-6445,-8233,-6799,-6513,-7375,-7807,-7727,-8093,-8211,-10813,-13501,-16946,-21548,-37132,},  // speed: 9
		{-4018,-7587,-10697,-7005,-6589,-7482,-5955,-6745,-7142,-7059,-7007,-5897,-7308,-7554,-8121,-8959,-9382,-11766,-20705,-29925,-39450,},  // speed: 10
		{-51636,-15350,-7186,-5716,-6705,-7359,-7047,-6979,-6384,-6909,-7435,-6951,-6906,-8147,-8123,-9130,-9776,-12918,-16742,-24616,-42586,},  // speed: 11
		{-27927,-5443,-11748,-9566,-5720,-7860,-6169,-8673,-8445,-7886,-7058,-9041,-7926,-8920,-9767,-10583,-11035,-13587,-18242,-13467,-35782,},  // speed: 12
		{-23167,-13335,-9950,-10681,-12635,-8809,-7193,-8186,-8276,-7671,-7649,-8101,-10224,-10237,-9013,-11550,-11721,-14124,-16037,-12663,-23816,},  // speed: 13
		{-44456,-3453,-13873,-9381,-9784,-10035,-10440,-11353,-6911,-9330,-8550,-10557,-11731,-14312,-11636,-12172,-13850,-16633,-22530,-8628,-32005,},  // speed: 14
		{-8778,-12590,-9084,-5900,-5992,-6134,-6873,-6838,-6944,-8363,-14966,-8003,-7833,-8612,-10797,-10447,-12852,-12614,-11039,-15686,-16098,},  // speed: 15
		{-5154,-16692,-16385,-11618,-8625,-10472,-7846,-7005,-10752,-7525,-6543,-7318,-12310,-11024,-11668,-11624,-15537,-12735,-15944,-13174,-37408,},  // speed: 16
		{-36691,-14784,-13059,-6655,-6188,-17797,-14152,-6411,-14357,-6832,-8874,-13757,-12202,-12294,-11919,-12650,-12585,-13529,-18583,-16684,-23158,},  // speed: 17
		{-5938,-17624,-16912,-27136,-6283,-10197,-7622,-9059,-11807,-14942,-4969,-12229,-6633,-21084,-12480,-12527,-12440,-12081,-37112,-17521,-21483,},  // speed: 18
		{-14379,-29888,-25876,-10471,-13811,-9396,-10926,-14846,-4893,-11713,-27664,-12766,-14434,-13455,-15179,-13981,-18158,-16589,-28956,-28882,-15959,},  // speed: 19
		{-9145,-4662,-13087,-12261,-16587,-8349,-9042,-8506,-13062,-13108,0,-19761,-12070,-11244,-20154,-17700,-18309,-21058,-23257,-40579,-21415,},  // speed: 20
		{-7781,-13723,-25937,-20040,-14650,-14567,-15719,-13989,-10286,-14628,-26999,-13406,-18491,-11364,-19153,-18171,-20956,-19703,-23828,-28270,-38020,},  // speed: 21
		{-28346,-25179,-14583,-11339,-18992,-13225,-14674,-17257,-17676,-21978,-14162,-18376,-17436,-15903,-20703,-19547,-21611,-24274,-28053,-24866,0,},  // speed: 22
		{-10000,-13251,-18429,-10050,-12047,-14512,-18095,-13139,-10333,-16142,-10823,-15102,-13480,-17987,-17635,-20414,-19579,-22597,-25882,-23142,-25839,},  // speed: 23
		{-17583,-22595,-9554,-17433,-11739,-13762,-15571,-16097,-15301,-16577,-22192,-16857,-15039,-17469,-17238,-19048,-19387,-21935,-23561,-20437,-18369,},  // speed: 24
		{-8749,-22098,-5542,-16801,-7161,-12948,-14357,-12540,-14405,-17150,-9679,-15448,-15904,-16302,-18293,-17658,-19816,-19352,-21475,-28257,-25441,},  // speed: 25
		{-34255,-21035,-15343,-19410,-16531,-14866,-15413,-16901,-11102,-17432,-20690,-16194,-16300,-14384,-16315,-17029,-16929,-18153,-19640,-23371,-21241,},  // speed: 26
		{-19142,-23562,0,-15425,-4034,-13650,-12979,-16135,-10207,-13170,0,-16504,-10899,-12494,-15044,-16601,-18649,-21064,-26167,-29210,-14605,},  // speed: 27
		{-25000,-24075,-27727,-22517,-4021,-25375,-10074,-12995,-18855,-11302,-14414,-14043,-12056,-13906,-12402,-14665,-23320,-23949,-22827,-19215,-36705,},  // speed: 28
		{-28703,-29964,-26502,-29281,-28269,-20307,-19264,-21140,-21313,-18817,-16346,-21535,-19686,-18229,-20221,-19503,-20803,-22455,-22590,-27001,-34009,},  // speed: 29
		{-27412,-42978,-25458,-24601,-24080,-22153,-18245,-21167,-18844,-22764,-19328,-21781,-20810,-25757,-19864,-24861,-24828,-27757,-26979,-26243,-38842,},  // speed: 30
		{-14875,-28165,-26554,-24552,-25744,-24042,-20146,-19469,-19974,-21169,-21507,-21101,-21111,-20546,-20714,-28274,-26829,-31097,-25321,-31995,-36249,},  // speed: 31
		{-25000,-25000,-25000,-21012,-23024,-25826,-18973,-24215,-21057,-21649,-19706,-19467,-27390,-18500,-20596,-21834,-21558,-22088,-26287,-33327,-47396,},  // speed: 32
		{-25000,-25000,-25000,-25000,-19089,-17385,-13709,-21901,-23169,-25716,-20172,-20840,-23252,-20252,-30091,-23230,-25839,-27003,-31162,-21981,-28584,},  // speed: 33
		{-25000,-25000,-18375,-15884,-17609,-17524,-22026,-13841,-15000,-15000,-15000,-18561,-16766,-24681,-20000,-20000,-24264,-28338,-29704,-26988,-20000,},  // speed: 34
	};
	private double[][] mikeBattPowerTargets = new double[][] {
		{-10000,-10000,-10000,-10000,-8095,-11494,-10456,-13652,-11954,-10561,-9989,-9940,-15043,-13652,-19256,-15734,-16609,-13873,-9856,-8728,-9818,},  // speed: 0
		{-10000,-10000,-10000,-17323,-13717,-16308,-14055,-14125,-14736,-15546,-14567,-15633,-15169,-19906,-19316,-21364,-20882,-29214,-16769,-17479,-12583,},  // speed: 1
		{-10000,-10000,-9233,-8403,-14186,-15115,-15496,-14627,-14186,-18454,-13752,-28579,-20091,-24821,-25124,-28106,-33758,-33376,-71444,-15210,-13279,},  // speed: 2
		{-15000,-15000,-5939,-12194,-13451,-12717,-12438,-15004,-15922,-14910,-13577,-15260,-18259,-18020,-19085,-26834,-29422,-41981,-33421,-13172,-12782,},  // speed: 3
		{-15000,-11241,-29276,-13403,-13317,-13957,-12898,-18299,-10694,-12918,-11583,-14359,-16011,-15178,-15110,-22984,-28761,-41073,-31767,-33520,-11871,},  // speed: 4
		{-15000,-6971,-9711,-12066,-11600,-13880,-11802,-10556,-15795,-13696,-18211,-15548,-15120,-21487,-20277,-19982,-25367,-35592,-37585,-84865,-13485,},  // speed: 5
		{-15000,-12890,-10770,-11148,-12210,-15180,-15511,-14258,-13691,-13821,-12421,-14683,-14545,-16272,-15015,-19124,-20662,-25784,-38016,-61570,-11217,},  // speed: 6
		{-20489,-10589,-12292,-13301,-13092,-14119,-13693,-13151,-12508,-14529,-13427,-13017,-17605,-16191,-16091,-15652,-18541,-22211,-38472,-58069,-29741,},  // speed: 7
		{-10613,-33903,-8058,-11388,-12581,-13343,-12600,-12206,-13245,-13357,-11186,-16608,-13569,-14312,-15040,-15617,-17564,-19708,-28384,-43773,-30924,},  // speed: 8
		{-22680,-11953,-12646,-11099,-11221,-11348,-13030,-13401,-13601,-12331,-13822,-11375,-13357,-13015,-13564,-15190,-16186,-17752,-26925,-48649,-28482,},  // speed: 9
		{-21862,-12348,-10469,-14678,-10910,-10630,-10922,-12673,-10678,-11156,-12198,-8806,-10956,-13004,-12860,-13715,-15034,-17324,-25821,-39290,-58159,},  // speed: 10
		{-42721,-18913,-17691,-11302,-12582,-10702,-11602,-10398,-15484,-11152,-10877,-11111,-11707,-11289,-13052,-12558,-14499,-15599,-21953,-31226,-42586,},  // speed: 11
		{-15506,-11456,-14542,-12064,-10618,-12096,-11016,-10817,-10128,-10487,-11891,-9867,-10875,-10631,-12058,-11796,-12038,-14361,-19012,-34131,-50724,},  // speed: 12
		{-35623,-16358,-16259,-13798,-11696,-11926,-11617,-12371,-11059,-11154,-10706,-13485,-15147,-11911,-12058,-11438,-11749,-14384,-16136,-18174,-23816,},  // speed: 13
		{-15822,-13046,-14564,-11824,-11732,-11240,-11331,-11106,-10649,-11125,-11118,-10832,-11336,-11693,-11939,-11746,-13209,-14297,-21579,-17321,-17967,},  // speed: 14
		{-7064,-9938,-11632,-13860,-12066,-8952,-10348,-8297,-9231,-10099,-10062,-11491,-8621,-9612,-10515,-10211,-11991,-14599,-15109,-14864,-25850,},  // speed: 15
		{-5154,-27538,-12270,-14260,-11075,-9385,-11702,-10423,-9314,-10859,-9486,-11988,-9828,-10451,-12372,-11811,-12212,-14733,-18968,-28716,-26777,},  // speed: 16
		{-3674,-27241,-21092,-6889,-14223,-12149,-12771,-15740,-11062,-10726,-7083,-11561,-18173,-14244,-16364,-13217,-16570,-15974,-18676,-31460,-30552,},  // speed: 17
		{-9163,0,-14973,-7102,-9292,-8825,-11716,-14195,-10864,-18498,-11311,-14620,-12955,-15621,-14592,-15731,-15964,-20965,-23307,-24823,-49350,},  // speed: 18
		{-14379,-53108,-7045,-9902,-8823,-11388,-10808,-15041,-14004,-14477,-12656,-15134,-13083,-14880,-14236,-17088,-18554,-18798,-23962,-22366,-15959,},  // speed: 19
		{-9145,-8400,0,-7104,-14016,-11988,-12390,-14664,-12208,-14181,-18657,-15176,-16996,-15744,-14187,-14523,-15618,-19542,-24943,-27697,0,},  // speed: 20
		{-3777,0,-837,-9289,-15500,-18568,-14251,-13876,-17347,-16813,-17081,-15115,-14934,-16556,-16994,-16411,-18189,-20877,-26155,-26926,-38020,},  // speed: 21
		{0,-2025,-9000,-14984,-16127,-15040,-13487,-15965,-13515,-19427,-17305,-17043,-18923,-16521,-17393,-19155,-19226,-21717,-24829,-24404,0,},  // speed: 22
		{-10000,-18705,-3897,-9682,-9612,-15536,-17767,-16391,-18023,-17013,-14439,-16871,-19370,-18785,-19893,-20419,-20996,-23422,-24385,-25152,-1545,},  // speed: 23
		{-17021,-18021,-7298,-10023,-20365,-16222,-17947,-18923,-19560,-20720,-15953,-19397,-21358,-21313,-20171,-22014,-24242,-24455,-25441,-26828,-18369,},  // speed: 24
		{-13511,-22098,-10582,-14271,-16255,-16749,-17433,-17558,-19919,-19073,-23693,-19312,-20333,-18817,-20465,-21670,-22337,-23797,-26545,-30952,-20856,},  // speed: 25
		{-34255,-30016,-12230,-16901,-14072,-18684,-19617,-18578,-17907,-17746,-18663,-17589,-19186,-18109,-19861,-19884,-20509,-22225,-24371,-27018,-33425,},  // speed: 26
		{-15883,-23562,-29613,-17568,-18481,-16743,-18965,-18689,-18662,-19629,-19999,-18629,-20095,-19111,-20089,-20356,-21207,-22223,-24483,-25194,-29796,},  // speed: 27
		{-25000,-24075,-25444,-21188,-20697,-20412,-20472,-20226,-20009,-20738,-18318,-21583,-20040,-20767,-20640,-20996,-21331,-21979,-23106,-26549,-28992,},  // speed: 28
		{-28703,-34723,-36127,-27876,-22304,-22764,-20510,-21903,-21305,-20749,-18948,-20830,-21036,-20484,-21145,-21061,-21537,-22104,-23420,-26270,-34540,},  // speed: 29
		{-17462,-38885,-32536,-31128,-23206,-22883,-29310,-22494,-22436,-23022,-23175,-26079,-21132,-21349,-22705,-23355,-22997,-24036,-26544,-31962,-33634,},  // speed: 30
		{-16481,-28089,-28842,-24552,-32581,-17103,-19984,-20705,-21274,-23358,-26132,-18209,-23616,-23177,-22507,-22540,-23141,-23530,-24902,-38157,-36249,},  // speed: 31
		{-25000,-25000,-25000,-36629,-23024,-25826,-17922,-24215,-21057,-19869,-19706,-19467,-27390,-18500,-20187,-21834,-17641,-21948,-26861,-33327,-65798,},  // speed: 32
		{-25000,-25000,-25000,-25000,-19089,-17385,-13709,-21901,-23169,-25716,-20172,-20840,-23252,-20252,-30091,-23230,-27785,-15197,-27081,-21981,-28584,},  // speed: 33
		{-25000,-25000,-18375,-15884,-17609,-17524,-22026,-13841,-15000,-15000,-15000,-18561,-16766,-24681,-15994,-20000,-24264,-28338,-29704,-26988,-20000,},  // speed: 34
	};
	private double[][] nikolayBattPowerTargets = new double[][] {
		{-10000,-10000,-10000,-10000,-15243,-8874,-14580,-9228,-9522,-10430,-9353,-8967,-9957,-14522,-13153,-14318,-15175,-13873,-9856,-8728,-9818,},  // speed: 0
		{-10000,-10000,-10000,-20430,-9696,-10310,-13235,-11190,-15427,-12222,-15069,-9434,-11492,-12959,-17007,-18100,-32738,-35482,-16769,-17479,-12583,},  // speed: 1
		{-10000,-10000,-10356,-7598,-16050,-10447,-11927,-13423,-11567,-18613,-11471,-15125,-16575,-17491,-16682,-25020,-17819,-35277,-16199,-15210,-13279,},  // speed: 2
		{-15000,-15000,-16650,-15922,-12166,-11832,-10610,-15021,-13305,-13644,-10449,-17230,-14038,-15175,-15195,-19153,-31105,-25016,-27019,-13172,-12782,},  // speed: 3
		{-15000,-8146,-17866,-9816,-12436,-11254,-15634,-11387,-12621,-11911,-15684,-14560,-13461,-16715,-15441,-18352,-24681,-28527,-29632,-19436,-11871,},  // speed: 4
		{-15000,-6971,-9694,-13170,-11861,-11410,-11589,-10369,-15286,-15344,-12530,-13243,-12562,-17679,-13662,-16504,-18915,-31131,-90656,-20863,-13485,},  // speed: 5
		{-15000,-8087,-11604,-11117,-10932,-9895,-9296,-10782,-11122,-11481,-14790,-12365,-10180,-11371,-11667,-16798,-17658,-24823,-29498,-27317,-11217,},  // speed: 6
		{-20489,-9362,-10470,-14149,-11889,-9072,-9687,-10487,-9451,-9979,-10068,-10263,-11597,-11654,-11843,-12106,-14150,-18241,-29298,-64741,-16376,},  // speed: 7
		{-8496,-15487,-9863,-13008,-9016,-11055,-9921,-9047,-9824,-9798,-10429,-12523,-10152,-10129,-11244,-11667,-12191,-16800,-25474,-49966,-104509,},  // speed: 8
		{-22680,-8107,-15601,-13826,-9537,-8869,-7803,-8878,-8576,-10346,-10780,-8699,-10309,-10379,-9900,-10527,-11474,-15747,-23886,-46402,-124249,},  // speed: 9
		{-21862,-21584,-11218,-10853,-8536,-8462,-8708,-8230,-8336,-9976,-10144,-8067,-11345,-10113,-8423,-9531,-10504,-14384,-23943,-48504,-15872,},  // speed: 10
		{-51636,-15091,-13001,-10490,-10454,-8820,-8746,-8478,-8019,-8728,-6819,-7740,-9085,-8330,-9269,-9521,-9707,-11325,-20323,-37848,-24235,},  // speed: 11
		{-45305,-11038,-13675,-10370,-8966,-8832,-7420,-8297,-8835,-8315,-9054,-7588,-8869,-8318,-8272,-8540,-9984,-12827,-15035,-37478,-42068,},  // speed: 12
		{-23167,-9767,-21517,-11191,-11345,-10798,-7524,-7912,-9123,-8971,-7715,-8644,-8596,-9299,-10254,-11073,-10183,-13042,-18077,-24617,-23816,},  // speed: 13
		{-6772,-4263,-13492,-13677,-8999,-12476,-10984,-9741,-9168,-8492,-9494,-8527,-9175,-11619,-11167,-10922,-12093,-12752,-19561,-13446,-33972,},  // speed: 14
		{-7064,-12590,-18093,-14882,-10836,-10338,-8852,-11091,-11518,-9719,-10297,-10414,-8400,-10805,-10438,-10451,-12327,-13158,-14922,-17682,-23470,},  // speed: 15
		{-5154,-14363,-14420,-13027,-9490,-11586,-11324,-12255,-12982,-10962,-13250,-11917,-11190,-10890,-12572,-12403,-12864,-14991,-20400,-18840,-33542,},  // speed: 16
		{-20297,-14947,-11823,-11343,-11330,-10499,-11311,-11434,-10977,-11159,-11480,-12321,-12997,-12115,-12641,-13099,-13785,-16529,-18320,-20464,-35348,},  // speed: 17
		{-9163,-11735,-7379,-9732,-11510,-11404,-11803,-12069,-12289,-12519,-11459,-12899,-14520,-13259,-13737,-13443,-14111,-15243,-17089,-19784,-41460,},  // speed: 18
		{-14379,-31640,-8737,-13021,-13736,-14479,-14512,-12341,-13453,-13730,-13314,-13902,-15475,-13890,-14097,-14125,-15382,-16486,-18679,-17794,-19892,},  // speed: 19
		{-24139,-18730,-24614,-19958,-15244,-14685,-17013,-13220,-13816,-14859,-13460,-14557,-15135,-15888,-14662,-15714,-16515,-17585,-19587,-20805,-5312,},  // speed: 20
		{-7781,-25258,-33472,-15390,-15659,-15094,-14400,-15963,-18236,-15964,-16677,-17755,-16450,-18289,-16902,-17868,-18252,-19647,-21436,-24860,-28230,},  // speed: 21
		{-28346,-10755,-21924,-15621,-15347,-15335,-15305,-15663,-17107,-16620,-16008,-15910,-16901,-17155,-17426,-17644,-17740,-18496,-21659,-33346,-22133,},  // speed: 22
		{-10087,-36681,-13946,-14501,-15855,-15596,-15597,-16418,-16194,-16526,-16251,-16556,-17189,-16994,-17510,-17413,-18385,-19017,-20660,-20232,-44254,},  // speed: 23
		{-17583,-22595,-14675,-17021,-16279,-18046,-15828,-16246,-19453,-17862,-16950,-19902,-18215,-17039,-17866,-18348,-19134,-20610,-21829,-23523,-26855,},  // speed: 24
		{-8749,-27398,-17484,-16531,-14949,-16270,-15754,-18034,-16377,-16292,-17959,-17632,-17774,-18479,-17275,-17828,-18668,-19570,-21254,-23967,-30305,},  // speed: 25
		{-34255,-9839,-14913,-16089,-15586,-13643,-16268,-16645,-17540,-16233,-17836,-15736,-16762,-16145,-16904,-18309,-20025,-20495,-20942,-26257,-33945,},  // speed: 26
		{-19142,-23562,-22059,-9274,-14602,-17471,-15967,-38286,-18680,-16854,-19551,-15880,-16240,-18151,-17527,-18433,-19507,-20560,-21436,-24884,-30687,},  // speed: 27
		{-25000,-24075,-19229,-9474,-16689,-13727,-16984,-18373,-15990,-19867,-11434,-18175,-25188,-17089,-19229,-20179,-20032,-21502,-20600,-27549,-15102,},  // speed: 28
		{-28703,-29964,-13392,-29281,-22943,-28436,-19264,-21140,-25984,-18899,-16346,-21535,-12017,-20911,-18606,-20401,-19022,-22412,-22940,-25321,-34009,},  // speed: 29
		{-27412,-42978,-25458,-24601,-24080,-24768,-18245,-21167,-18844,-18867,-19328,-21781,-20810,-21137,-22523,-20045,-29693,-27755,-25007,-28468,-21718,},  // speed: 30
		{-14875,-28165,-26554,-24552,-25744,-21843,-20146,-19469,-19974,-21169,-21507,-21101,-21111,-20546,-20714,-22554,-23445,-24139,-25321,-31995,-36249,},  // speed: 31
		{-25000,-25000,-25000,-36629,-23024,-25826,-18973,-24215,-21057,-21649,-19706,-19467,-27390,-18500,-20596,-21834,-21558,-23919,-26423,-33327,-47396,},  // speed: 32
		{-25000,-25000,-25000,-25000,-19089,-17385,-13709,-21901,-23169,-25716,-20172,-20840,-23252,-20252,-30091,-23230,-25839,-27003,-31162,-21981,-28584,},  // speed: 33
		{-25000,-25000,-18375,-15884,-17609,-17524,-22026,-13841,-15000,-15000,-15000,-18561,-16766,-24681,-20000,-20000,-24264,-28338,-29704,-26988,-20000,},  // speed: 34
	};
	private double [][] richBattPowerTargets = new double[][] {
		{-10000,-10000,-10000,-10000,-15243,-14144,-5155,-10665,-11996,-9382,-9473,-7660,-7508,-8138,-8990,-9629,-7770,-13873,-9856,-8728,-9818,},  // speed: 0
		{-10000,-10000,-10000,-5546,-5780,-13794,-9600,-6610,-11490,-7723,-7416,-8493,-9627,-12889,-13354,-9979,-8511,-7887,-22056,-17479,-12583,},  // speed: 1
		{-10000,-10000,-6567,-9447,-23097,-8904,-6598,-7496,-7605,-7048,-6918,-7791,-9010,-10110,-11761,-13275,-9710,-12238,-11394,-15210,-13279,},  // speed: 2
		{-15000,-4208,-8219,-7141,-5985,-8992,-7294,-6417,-6253,-6887,-7698,-7836,-7912,-8177,-9677,-12788,-14077,-13765,-25537,-33136,-12782,},  // speed: 3
		{-15000,-5938,-4809,-9684,-6339,-12040,-7254,-6537,-7176,-6795,-6430,-7488,-9326,-8223,-8972,-12202,-15396,-18265,-23309,-34080,-11871,},  // speed: 4
		{-4776,-23499,-6577,-7602,-6953,-6773,-7050,-6862,-6791,-7089,-8728,-7858,-8674,-8177,-8881,-10487,-12939,-18926,-13139,-10223,-63324,},  // speed: 5
		{-5320,-71014,-14443,-8711,-7260,-7658,-8134,-7485,-6687,-16732,-31376,-7864,-9044,-8406,-12675,-10454,-13427,-18616,-32200,-9595,-25733,},  // speed: 6
		{-5894,-3681,-11823,-8617,-8599,-7012,-7077,-8257,-7280,-7544,-9042,-6745,-8101,-8798,-9971,-10484,-11417,-14774,-11145,-7727,-16376,},  // speed: 7
		{-5063,-4023,-16833,-7025,-6175,-6274,-7771,-6493,-6477,-6890,-6907,-7021,-7385,-7802,-7590,-9612,-10460,-15069,-15401,-8038,-30924,},  // speed: 8
		{-22680,-51102,-20594,-8795,-6934,-8822,-6952,-6244,-6531,-6685,-6217,-7352,-7056,-7584,-6896,-8145,-10539,-13155,-16201,-28691,-46448,},  // speed: 9
		{-21862,-6899,-20290,-8913,-7134,-6680,-7352,-6054,-6000,-5939,-5636,-5322,-6374,-7344,-8017,-8364,-10431,-12939,-22789,-30166,-44539,},  // speed: 10
		{-51636,-28984,-5998,-5078,-5614,-8011,-7046,-6570,-6268,-6100,-6463,-6145,-7016,-7910,-7907,-7919,-9555,-12284,-22113,-13214,-42586,},  // speed: 11
		{-27927,-16425,-2915,-4611,-6245,-7459,-7958,-5553,-5953,-8778,-5251,-5888,-12449,-10471,-11902,-7413,-8390,-12216,-17853,-14763,-34525,},  // speed: 12
		{-5291,-5484,-9352,-9317,-13917,-6117,-8819,-5426,-7326,-8556,-6082,-6641,-5772,-9508,-6746,-7792,-8815,-13427,-16205,-26228,-28690,},  // speed: 13
		{-11155,-37322,-4493,-37728,-19039,-7857,-6023,-5820,-5394,-6130,-7067,-6171,-6182,-7632,-6661,-7620,-9133,-13226,-20729,-11131,-101723,},  // speed: 14
		{-7064,-85359,-2565,-13526,-6989,-5111,-5610,-9313,-6571,-6764,-6024,-7174,-7542,-5952,-7408,-7165,-8975,-13543,-27146,-28818,-20938,},  // speed: 15
		{-5154,-20506,-542,-7659,-8633,-8405,-6139,-6249,-8788,-6496,-11473,-6836,-6958,-7450,-7520,-8310,-9928,-15691,-21876,-7554,-9000,},  // speed: 16
		{-14476,-19442,-6927,-16224,-6067,-6231,-6162,-5688,-5632,-5382,-6080,-5967,-6607,-6289,-7719,-8781,-10451,-16201,-26791,-6082,-23141,},  // speed: 17
		{-9163,-17624,-916,-4925,-8279,-9931,-7189,-8001,-8234,-10126,-5701,-9915,-7986,-7174,-8328,-9368,-10266,-16317,-27754,-64605,-37347,},  // speed: 18
		{-14379,-35546,-9552,-8319,-17103,-9531,-8516,-8742,-7181,-7016,-10511,-9921,-6680,-9682,-10112,-10456,-12104,-16512,-30522,-12430,-26046,},  // speed: 19
		{-9145,-4662,-10342,-12483,-7749,-10698,-9485,-8803,-6311,-9205,-8618,-11118,-11887,-9480,-11205,-10395,-13872,-17357,-27337,-23263,-5312,},  // speed: 20
		{-7781,-6656,-8576,-4757,-9776,-10308,-11653,-8257,-8813,-11184,-5223,-12072,-11438,-10262,-12962,-13379,-14633,-18287,-21609,-35333,-38020,},  // speed: 21
		{-28346,-38632,0,-9984,-7753,-12228,-12260,-10517,-14694,-13131,-12478,-11371,-13499,-11135,-12801,-12740,-14804,-18817,-20927,-22822,0,},  // speed: 22
		{-10000,-13251,-9395,-8164,-8280,-12447,-8646,-12765,-16026,-12361,-12621,-12609,-11841,-14529,-15973,-15604,-18052,-20461,-23499,-15331,-25839,},  // speed: 23
		{-17583,-22595,-9554,-8465,-18531,-17958,-13910,-14387,-11261,-15034,-14072,-9656,-14686,-15597,-16510,-16490,-17764,-19604,-24149,-37661,-18369,},  // speed: 24
		{-8749,-22098,-19252,-10415,-14114,-20721,-16649,-14544,-16920,-13732,-17382,-15359,-18091,-16672,-15873,-16112,-18459,-19577,-23729,-30354,-25441,},  // speed: 25
		{-34255,-21035,-12923,-5762,-17666,-14203,-15810,-17384,-12008,-16432,-17314,-13791,-16633,-16987,-15830,-16230,-18619,-21319,-24899,-21540,-57164,},  // speed: 26
		{-19142,-23562,-26188,-15965,-8257,-19551,-13733,-17841,-16240,-17073,-19018,-18014,-17303,-18637,-16754,-18115,-19926,-22970,-26267,-25469,-40155,},  // speed: 27
		{-10754,-24075,-11679,-15304,0,-10746,-16739,-14179,-15557,-15979,-18915,-15161,-15749,-18897,-19676,-19409,-20354,-22681,-25399,-29457,-36705,},  // speed: 28
		{-28703,-29964,-28961,-29281,-21742,-16032,-11106,-14525,-21181,-15908,-19961,-16428,-18202,-19282,-18353,-20103,-21682,-23776,-26822,-30510,-34009,},  // speed: 29
		{-27412,-42978,-25458,-24601,-18479,-21107,-19858,-12839,-16471,-14810,-20073,-16225,-17184,-16825,-17122,-18930,-20043,-22083,-25850,-27300,-38842,},  // speed: 30
		{-14875,-28165,-26554,-24552,-25744,-21843,-15781,-14388,-16394,-15673,-15312,-17142,-14358,-15782,-16059,-16878,-18107,-18816,-21036,-31995,-36249,},  // speed: 31
		{-25000,-25000,-25000,-36629,-23024,-25826,-14702,-24215,-23320,-15650,-16140,-19467,-16896,-15330,-15759,-15144,-17838,-17050,-25969,-33327,-47396,},  // speed: 32
		{-25000,-25000,-25000,-25000,-19089,-17385,-13709,-21901,-23169,-25716,-20172,-20840,-23252,-20252,-30091,-14305,-19939,-20774,-21819,-21981,-28584,},  // speed: 33
		{-25000,-25000,-18375,-15884,-17609,-17524,-22026,-13841,-15000,-15000,-15000,-18561,-16766,-24681,-20000,-20000,-24264,-28338,-29704,-26988,-20000,},  // speed: 34
	};
	private double[][] thorBattPowerTargets = new double[][] {
		{-10000,-10000,-10000,-10000,-15243,-7918,-8270,-8005,-7880,-8335,-8303,-7881,-8993,-9157,-9711,-10979,-12323,-13873,-9856,-8728,-9818,},  // speed: 0
		{-10000,-10000,-10000,-5129,-8323,-8508,-8354,-8044,-9212,-9416,-8864,-9502,-10435,-11574,-10824,-11944,-14199,-13801,-16769,-17479,-12583,},  // speed: 1
		{-10000,-10000,-8385,-5781,-7061,-8598,-7854,-9067,-9526,-9222,-10379,-9746,-11497,-11787,-13201,-15725,-19586,-31369,-18519,-15210,-13279,},  // speed: 2
		{-15000,-15000,-16650,-6246,-8300,-8647,-8088,-9222,-8651,-9067,-10383,-10883,-9971,-11523,-12446,-17013,-20832,-24417,-52922,-13172,-12782,},  // speed: 3
		{-15000,-8146,-3471,-15016,-7342,-7907,-8298,-8775,-8178,-10068,-8852,-9463,-10475,-12532,-12638,-14166,-18618,-25247,-37619,-19436,-11871,},  // speed: 4
		{-15000,-6971,-6915,-5970,-8970,-7885,-10287,-9528,-8739,-9403,-10220,-19118,-9782,-11337,-12596,-12405,-17647,-23079,-30131,-20215,-13485,},  // speed: 5
		{-15000,-8057,-7333,-6219,-7407,-7912,-9438,-8693,-9830,-9272,-7722,-10289,-8541,-12987,-12160,-13357,-14408,-21688,-21520,-24254,-11217,},  // speed: 6
		{-20489,-9642,-16303,-6827,-7774,-7765,-7641,-7942,-8613,-9806,-15097,-9632,-9684,-9408,-11100,-11603,-12797,-19475,-27434,-38403,-8860,},  // speed: 7
		{-8602,-46580,-14380,-6399,-6713,-8456,-10247,-8500,-8368,-9214,-12405,-11095,-9574,-10766,-11672,-10900,-13373,-16964,-25891,-47192,-76153,},  // speed: 8
		{-22680,-17327,-14909,-8901,-8881,-9228,-9576,-10536,-8005,-9397,-11823,-13275,-10142,-9765,-10391,-10812,-12682,-16949,-23147,-37600,-11661,},  // speed: 9
		{-41202,-36850,-19324,-16640,-8745,-10124,-9885,-9352,-9965,-9017,-10962,-10078,-8925,-10830,-9888,-10582,-13195,-14850,-23912,-36195,-59705,},  // speed: 10
		{-58725,-25404,-25162,-12086,-10437,-9331,-10008,-9544,-11883,-10815,-8511,-10411,-11207,-12694,-11076,-11430,-13386,-17149,-23257,-37640,-42586,},  // speed: 11
		{-17121,-16768,-14065,-10772,-12744,-10288,-9120,-8363,-11114,-10093,-11537,-10410,-10690,-11077,-11399,-11502,-13853,-15963,-23965,-35274,-20836,},  // speed: 12
		{-21302,-16021,-11432,-11041,-11375,-10065,-8964,-9489,-9048,-9888,-9208,-13651,-9394,-13477,-11766,-11640,-12923,-16436,-22648,-21273,-23816,},  // speed: 13
		{-17796,-6548,-6619,-8573,-9845,-9370,-10282,-10876,-12033,-10576,-12000,-11930,-11214,-12207,-12875,-13040,-14024,-15811,-16781,-15619,-46474,},  // speed: 14
		{-9504,-16693,-12104,-9985,-9353,-9212,-10398,-10698,-11611,-11894,-13341,-12814,-13929,-12678,-12584,-12848,-14890,-14196,-16558,-17532,-25516,},  // speed: 15
		{-5335,-15837,-5918,-10267,-8357,-8695,-8574,-9041,-8615,-9912,-11298,-10883,-10184,-10784,-11897,-13771,-14650,-16286,-17814,-14306,-25953,},  // speed: 16
		{-36691,-41218,-3362,-11316,-9191,-7579,-8453,-9446,-8787,-7934,-8187,-9809,-9532,-10541,-10488,-12702,-12477,-14494,-20647,-21883,-19581,},  // speed: 17
		{-25124,-2617,-7116,-8267,-7669,-8322,-8681,-9075,-11513,-9792,-9073,-11791,-9982,-10442,-10893,-12895,-14260,-14841,-15401,-26295,-66357,},  // speed: 18
		{-11688,-19114,-12173,-9296,-8006,-10412,-9012,-8494,-9913,-9959,-11984,-8880,-9717,-11652,-11295,-12822,-14003,-15604,-18743,-17924,-24620,},  // speed: 19
		{-14355,-24853,-12811,-13390,-10021,-10612,-12011,-10370,-12488,-10792,-12508,-11513,-13734,-11220,-12258,-13862,-16717,-18071,-20660,-21478,-17997,},  // speed: 20
		{-7781,-24362,-6079,-13645,-10026,-10531,-10462,-10292,-11182,-11137,-13607,-11570,-11053,-9655,-14461,-13891,-15572,-16721,-20794,-44334,-50651,},  // speed: 21
		{-16881,-21695,-34822,-17375,-11983,-11304,-10891,-11526,-12538,-11215,-12968,-10284,-12119,-12513,-11879,-13909,-13037,-14980,-15781,-25334,-817,},  // speed: 22
		{-5191,-18308,-7306,-24193,-17725,-10281,-18673,-18657,-12543,-14156,-15225,-11974,-16136,-13323,-17862,-15622,-19580,-17727,-20581,-29427,-21958,},  // speed: 23
		{-17583,-19929,-12191,-29806,-16712,-32696,-10530,-24005,-9467,-18033,-16518,-15515,-19956,-10186,-18336,-19320,-26581,-27667,-24948,-22738,-18369,},  // speed: 24
		{-8749,-22098,-36147,-7463,-7745,-16711,-16855,-16964,-17908,-7530,-17908,-16859,-7184,-18789,-19249,-19924,-14475,-7746,-26193,-31619,-19757,},  // speed: 25
		{-34255,-21035,-23652,-17706,-18777,-18283,-17799,-18737,-17142,-17633,-18518,-17865,-18276,-17507,-20062,-19734,-19556,-21700,-24062,-27796,-29065,},  // speed: 26
		{-19142,-23562,-26188,-20382,-18871,-17732,-19142,-20219,-17132,-19861,-19551,-19477,-19396,-18864,-19781,-19525,-21499,-16767,-24687,-28044,-40155,},  // speed: 27
		{-25000,-24075,-27727,-22517,-18809,-18390,-19331,-18508,-18855,-20971,-17760,-20488,-19470,-19924,-21531,-20589,-20307,-22322,-23094,-29457,-36705,},  // speed: 28
		{-28703,-29964,-26502,-29281,-20882,-20307,-19264,-21140,-21313,-18817,-16346,-21535,-18403,-18611,-20221,-20605,-21840,-21903,-23931,-27001,-34009,},  // speed: 29
		{-27412,-42978,-25458,-24601,-24080,-21809,-18245,-21167,-18844,-18867,-19328,-21781,-20810,-21137,-22050,-20045,-22643,-23584,-25007,-30747,-38842,},  // speed: 30
		{-14875,-28165,-26554,-24552,-25744,-21843,-20146,-19469,-19974,-21169,-21507,-21101,-21111,-20546,-20714,-22554,-23445,-24139,-25321,-31995,-36249,},  // speed: 31
		{-25000,-25000,-25000,-36629,-23024,-25826,-18973,-24215,-21057,-21649,-19706,-19467,-27390,-18500,-20596,-21834,-21558,-23919,-26423,-33327,-47396,},  // speed: 32
		{-25000,-25000,-25000,-25000,-19089,-17385,-13709,-21901,-23169,-25716,-20172,-20840,-23252,-20252,-30091,-23230,-25839,-27003,-31162,-21981,-28584,},  // speed: 33
		{-25000,-25000,-18375,-15884,-17609,-17524,-22026,-13841,-15000,-15000,-15000,-18561,-16766,-24681,-20000,-20000,-24264,-28338,-29704,-26988,-20000,},  // speed: 34
	};
	private double[][] defaultBattTargetWatts = new double[][]{
		{-15000,-15000,-15000,-15000,-13312,-9043,-9530,-10333,-12410,-8762,-8513,-8996,-12360,-11104,-12719,-12561,-14521,-13873,-15000,-15000,-15000,}, // speed: 0
		{-15000,-15000,-15000,-20727,-11232,-11779,-10923,-12460,-11857,-12169,-12094,-11366,-12311,-14094,-14423,-16232,-16459,-16938,-18571,-20000,-20000,}, // speed: 1
		{-15000,-15000,-8715,-11630,-12655,-11422,-11071,-11401,-11980,-12280,-12481,-13322,-13421,-15560,-16227,-19805,-20894,-24009,-27088,-15210,-20000,}, // speed: 2
		{-15000,-15000,-17108,-12878,-11672,-10992,-11070,-11800,-11869,-13085,-12545,-12857,-14427,-15626,-16140,-19471,-22316,-25330,-26155,-13172,-25000,}, // speed: 3
		{-15000,-8401,-20677,-14037,-10213,-10712,-10678,-12019,-10576,-11493,-11001,-11911,-12785,-13525,-14703,-16969,-21001,-27492,-24867,-19635,-30000,}, // speed: 4
		{-15000,-7696,-12890,-11217,-10276,-9796,-10307,-11176,-12830,-11314,-12819,-14523,-11749,-13915,-13848,-15694,-18656,-24674,-30311,-20470,-35000,}, // speed: 5
		{-15000,-9971,-12254,-11375,-10051,-9633,-10533,-10046,-10718,-11433,-11941,-11289,-12082,-12879,-13495,-14647,-15790,-21246,-30904,-35605,-40000,}, // speed: 6
		{-20566,-10293,-11898,-11418,-9550,-8992,-9569,-9948,-9317,-10518,-11506,-10062,-11594,-11062,-12301,-12707,-14830,-18942,-29286,-41722,-16103,}, // speed: 7
		{-8683,-34674,-15583,-11059,-9604,-9297,-9897,-9262,-9374,-10255,-10183,-11018,-10056,-11640,-11185,-12049,-14281,-17360,-25357,-38235,-60398,}, // speed: 8
		{-22421,-18598,-15651,-10158,-9587,-9115,-9270,-9665,-9338,-9235,-10887,-9160,-10013,-10441,-10482,-11296,-13294,-16180,-23662,-43801,-48922,}, // speed: 9
		{-27049,-30749,-16752,-12018,-9309,-9488,-8906,-8756,-9040,-9721,-9466,-8975,-9770,-10049,-10224,-10922,-12586,-15076,-22686,-39132,-39056,}, // speed: 10
		{-51933,-28210,-18586,-12116,-9604,-8949,-9312,-9114,-10165,-9171,-10624,-8936,-9516,-10385,-10098,-10448,-11597,-14492,-21548,-31358,-51473,}, // speed: 11
		{-27422,-16721,-14485,-10588,-9938,-9814,-9085,-9335,-9218,-9107,-9869,-8934,-9817,-9601,-10072,-10788,-12511,-14651,-20376,-32941,-66180,}, // speed: 12
		{-23494,-16755,-14636,-12891,-10649,-9803,-9740,-9678,-9594,-9412,-9009,-10084,-10104,-9987,-10616,-11193,-12242,-15086,-19494,-24752,-28551,}, // speed: 13
		{-13330,-13410,-11574,-11551,-9903,-10324,-9848,-10084,-9955,-10271,-9683,-10226,-10342,-11135,-11257,-11838,-12924,-14745,-19874,-23699,-32656,}, // speed: 14
		{-8285,-12071,-12398,-12155,-9466,-9267,-9228,-9077,-9882,-9974,-10183,-9808,-10062,-10355,-10567,-11545,-12282,-14324,-16927,-17314,-25936,}, // speed: 15
		{-5215,-17015,-15524,-10526,-9615,-8894,-9210,-9554,-9178,-10153,-9126,-10464,-9825,-10575,-10950,-11250,-12332,-14428,-19303,-23550,-25669,}, // speed: 16
		{-38137,-20068,-12995,-9099,-10163,-8965,-9029,-9922,-9372,-9268,-10299,-9592,-10960,-10874,-10961,-11652,-12528,-14838,-18958,-24935,-31661,}, // speed: 17
		{-9451,-19049,-9321,-8765,-8892,-8824,-10252,-9670,-9656,-10927,-9621,-10656,-11010,-10744,-11324,-12223,-13127,-15785,-19874,-29877,-32180,}, // speed: 18
		{-13565,-36751,-10063,-10260,-9741,-9741,-10763,-10483,-11715,-11397,-12037,-11374,-12143,-11856,-12227,-13503,-14915,-16283,-20563,-20764,-18375,}, // speed: 19
		{-7898,-2157,-12962,-10570,-11854,-10632,-10748,-12382,-11486,-12529,-13454,-12859,-13129,-12854,-13952,-14659,-15910,-17996,-21249,-23048,-459,}, // speed: 20
		{-1391,-6310,-7729,-11797,-14112,-13329,-11997,-12619,-14613,-13986,-14556,-14577,-13419,-14174,-15077,-16469,-17907,-20146,-22789,-24778,-45730,}, // speed: 21
		{-27264,-620,-8117,-19115,-15688,-13877,-13268,-14249,-14923,-15746,-15490,-14483,-16283,-17834,-17518,-18312,-18580,-20805,-23180,-24970, 0,}, // speed: 22
		{-20000,-17120,-9871,-15046,-14040,-14872,-16257,-16200,-15797,-15616,-15342,-14628,-18208,-18138,-17700,-18718,-19949,-21896,-23472,-23029,-25060,}, // speed: 23
		{-20523,-20677,-10807,-18042,-17635,-15213,-15865,-16707,-17864,-17671,-15501,-19455,-19040,-17847,-18829,-19748,-21024,-22647,-25179,-29643,-30172,}, // speed: 24
		{-7215,-21355,-20247,-17671,-16700,-16329,-16949,-17293,-17595,-18116,-18284,-17869,-18412,-18628,-18524,-19985,-20821,-22850,-25488,-30844,-36716,}, // speed: 25
		{-37593,-25126,-21641,-20599,-16512,-18568,-18044,-18772,-18215,-18008,-18513,-18949,-18691,-18747,-19667,-20235,-21216,-23132,-24614,-28617,-37691,}, // speed: 26
		{-22124,-23156,-26338,-22745,-18984,-18207,-19253,-19490,-18017,-19377,-20121,-19345,-19767,-19472,-20126,-20634,-21604,-23209,-25057,-29530,-36699,}, // speed: 27
		{-25000,-21711,-27163,-21781,-20823,-20045,-19752,-19730,-19453,-20354,-19066,-21021,-20202,-20726,-20781,-21162,-21683,-22688,-24105,-29130,-36238,}, // speed: 28
		{-26321,-29345,-27703,-28824,-22781,-20359,-19892,-21229,-21150,-20206,-18742,-20979,-20673,-20252,-21005,-20906,-21866,-22873,-24215,-28205,-33177,}, // speed: 29
		{-29839,-39413,-29060,-23928,-24906,-22468,-20260,-21601,-20640,-20403,-20402,-21606,-20839,-20742,-21145,-21595,-22471,-23902,-25639,-32276,-35788,}, // speed: 30
		{-16129,-26510,-26977,-24067,-26790,-21824,-20529,-19079,-19516,-21537,-18866,-22253,-22207,-20363,-21012,-21493,-23155,-24716,-26091,-31078,-39178,}, // speed: 31
		{-30000,-30000,-25000,-36757,-22481,-24646,-19240,-24239,-20446,-21363,-18697,-19422,-27719,-18410,-20605,-22477,-22316,-24938,-26789,-31352,-46309,}, // speed: 32
		{-30000,-30000,-25000,-25000,-17432,-17140,-13348,-21364,-23382,-25011,-20132,-21727,-39376,-19802,-33054,-22944,-25135,-26408,-29778,-31263,-30853,}, // speed: 33
		{-30000,-30000,-18080,-16274,-20000,-17490,-24235,-14378,-20000,-20000,-20000,-20000,-16373,-23771,-19886,-18828,-23935,-28077,-29372,-26922,-30000,}, // speed: 34
	};
	private double[][] battPowerTargets = null;

	public void loadState() {
		alikPointTargetMap = loadPointTargets("/pointtargets/alikpointtargets.txt");
		arnoldPointTargetMap = loadPointTargets("/pointtargets/arnoldpointtargets.txt");
		garyPointTargetMap = loadPointTargets("/pointtargets/garypointtargets.txt");
		illahPointTargetMap = loadPointTargets("/pointtargets/illahpointtargets.txt");
		mikePointTargetMap = loadPointTargets("/pointtargets/mikepointtargets.txt");
		nikolayPointTargetMap = loadPointTargets("/pointtargets/nikolaypointtargets.txt");
		richPointTargetMap = loadPointTargets("/pointtargets/richpointtargets.txt");
		thorPointTargetMap = loadPointTargets("/pointtargets/thorpointtargets.txt");
	}

	/**
	 * Return key-value pairs in resource files as a Map.
	 */
	private static Map<String,Double> loadPointTargets(String resourceName) {

		Map<String,Double> ptMap = new HashMap<String,Double>(26000);
		try {
			InputStream is = MPL1031PolicyV5.class.getResourceAsStream(resourceName);
			BufferedReader br = new BufferedReader(new InputStreamReader(is));
			while (true) {
				String textLine = br.readLine();
				if (textLine == null) {
					break;
				}
				String[] data = textLine.split(" ");
				Double targetWattsDouble = Double.parseDouble(data[1]);
				ptMap.put(data[0], targetWattsDouble);
			}
		}
		catch (Exception e) {
			throw new RuntimeException("can't read from resource: " + resourceName + 
					"; please mount MPL1031PointTargets.jar on classpath");
		}
		return ptMap;
	}

	public void beginTrip(TripFeatures tripFeatures, BatteryModel batteryClone, BatteryModel capacitorClone) {
		modelBatt = batteryClone;
		modelCap = capacitorClone;
		capTotalWattsCapacity = modelCap.getMaxPower(1000);
		priorLatitude = priorLongitude = UNKNOWN;
		priorLat = priorLon = UNKNOWN;
		compassPoint = "";
		
		// get battPowerTargets and pointTargetMap for individual driver
		String driver = tripFeatures.getDriver().toLowerCase();
		if (driver.equalsIgnoreCase("alik")) {
			battPowerTargets = alikBattPowerTargets;
			pointTargetMap = alikPointTargetMap;
		}
		else if (driver.equalsIgnoreCase("arnold")) {
			battPowerTargets = arnoldBattPowerTargets;
			pointTargetMap = arnoldPointTargetMap;
		}
		else if (driver.equalsIgnoreCase("gary")) {
			battPowerTargets = garyBattPowerTargets;
			pointTargetMap = garyPointTargetMap;
		}
		else if (driver.equalsIgnoreCase("illah")) {
			battPowerTargets = illahBattPowerTargets;
			pointTargetMap = illahPointTargetMap;
		}
		else if (driver.equalsIgnoreCase("mike")) {
			battPowerTargets = mikeBattPowerTargets;
			pointTargetMap = mikePointTargetMap;
		}
		else if (driver.equalsIgnoreCase("nikolay")) {
			battPowerTargets = nikolayBattPowerTargets;
			pointTargetMap = nikolayPointTargetMap;
		}
		else if (driver.equalsIgnoreCase("rich")) {
			battPowerTargets = richBattPowerTargets;
			pointTargetMap = richPointTargetMap;
		}
		else if (driver.equalsIgnoreCase("thor")) {
			battPowerTargets = thorBattPowerTargets;
			pointTargetMap = thorPointTargetMap;
		}
		else {  // (default)
			battPowerTargets = defaultBattTargetWatts;
			pointTargetMap = null;
		}
	}

	public PowerFlows calculatePowerFlows(PointFeatures pf) {

		final int periodMS = pf.getPeriodMS(); // always 1000 as of 9/1/2010 change to TripBuilder.interpolatePoints()
		final double motorWattsDemandedOrGenerated = pf.getPowerDemand();
		final double speed = pf.getSpeed();
		final double minusCapWattsAvailable = modelCap.getMinPower(periodMS);
		final double plusCapWattsToFull = modelCap.getMaxPower(periodMS);

		final double latitude = pf.getLatitude();
		final double longitude = pf.getLongitude();
		final double lat = roundOff(latitude, 5000.0);
		final double lon = roundOff(longitude, 5000.0);

		// at a new PointTarget?
		if (lat != priorLat || lon != priorLon) {
			// past initial PointTarget?
			if (priorLatitude != UNKNOWN) {
				compassPoint = getCompassPoint(priorLatitude, priorLongitude, latitude, longitude);
			}
			// look for existing PointTarget in map
			String key = lat + "x" + lon + "y" + compassPoint;
			Double targetWattsDouble = null;
			// map is available?
			if (pointTargetMap != null) {
				targetWattsDouble = pointTargetMap.get(key);
			}
			// target was found?
			if (targetWattsDouble != null) {
				targetWatts = targetWattsDouble.doubleValue();
			}
			else {  // (PointTarget wasn't found)
				// speed in meters/sec up to 34 (34 meters/sec = 76.06 mph)
				int speedIndex = Math.min((int) Math.floor(speed), 34);  // for battPowerTargets[35][n]
				// minWattsUsed: -342,077.32  Math.log() = -12.74
				// maxWattsUsed: 67,171.35  Math.log() = 11.12
				// Math.max() eliminate negative logs betwen 0 and 1
				double absWattsLog = Math.log(Math.max(1, Math.abs(motorWattsDemandedOrGenerated)));
				// eliminate logs from -6 to +6 (+/-400 watts) as insignificant, range is now -6.74 to +5.12,
				//  double log to increase sensitivity; range is now -13.48 to +10.24
				absWattsLog = Math.max(absWattsLog - 6.0, 0.0) * 2.0;
				// restore sign, leading minus (negate) makes acceleration positive and braking negative
				double wattsLog = -Math.copySign(absWattsLog, motorWattsDemandedOrGenerated);
				// min/max eliminates a few extreme data points, offset so wattsLog 0 is at index 9
				int wattsIndex = Math.max(0, Math.min(20, (int) Math.round(wattsLog) + 9));  // 0-20: BBB...bbb_aaa...AAA
				targetWatts = battPowerTargets[speedIndex][wattsIndex];
			}
			// set aside unrounded latitude and longitude for direction of next PointTarget
			priorLatitude = latitude;
			priorLongitude = longitude;
		}
		// set aside rounded latitude and longitude to see if PointTarget changes
		priorLat = lat;
		priorLon = lon;

		// adjust target based on cap usage (force non-negative when plusCapWattsToFull is < 0)
		double battTargetWatts = targetWatts * Math.sqrt(Math.max(plusCapWattsToFull, 0.0) / capTotalWattsCapacity);

		//-------------------------------------------
		double battWattsToFromMotor = 0.0;
		double capWattsToFromMotor = 0.0;
		double battWattsToCap = 0.0;  // (cap never charges battery)

		// motor needs power (-)?
		if (motorWattsDemandedOrGenerated < 0.0) {
			// motor needs more than the battery power target?
			if (motorWattsDemandedOrGenerated < battTargetWatts) {
				// cap provides (-) power beyond battery target up to cap capacity
				capWattsToFromMotor = Math.max(motorWattsDemandedOrGenerated - battTargetWatts, minusCapWattsAvailable);
				// battery provides (-) the rest of the power (battery target or more if cap is empty)
				battWattsToFromMotor = motorWattsDemandedOrGenerated - capWattsToFromMotor;
			}
			else {  // (motor needs same or less than the battery power target)
				// battery provides (-) all power needed by motor
				battWattsToFromMotor = motorWattsDemandedOrGenerated;
				// motor needs less than the battery power target?
				if (motorWattsDemandedOrGenerated > battTargetWatts) {
					// battery also charges (-) cap, limit to rest of battery target or when cap full
					battWattsToCap = Math.max(battTargetWatts - motorWattsDemandedOrGenerated, -plusCapWattsToFull);
				}
			}
		}
		// motor is generating power (+)?
		else if (motorWattsDemandedOrGenerated > 0.0) {
			// motor is generating more power than the cap can store?
			if (motorWattsDemandedOrGenerated > plusCapWattsToFull) {
				// motor fully charges (+) cap
				capWattsToFromMotor = plusCapWattsToFull;
				// the rest charges (+) the battery
				battWattsToFromMotor = motorWattsDemandedOrGenerated - capWattsToFromMotor;
			}
			else {  // (motor generating same or less than the cap can store)
				// motor charges (+) cap
				capWattsToFromMotor = motorWattsDemandedOrGenerated;
				// cap not full?
				if (motorWattsDemandedOrGenerated < plusCapWattsToFull) {
					// battery also charges (-) cap, limit to battery target or when cap full
					battWattsToCap = Math.max(battTargetWatts, motorWattsDemandedOrGenerated - plusCapWattsToFull);
				}
			}
		}
		// (else) motor is not demanding or generating power

		//------------------------------------------
		// (DON'T CHANGE THIS) keep track of power flows on the cap and battery clones
		//  (duplicates the work actually done in the Simulator class)
		try {
			modelCap.drawPower(capWattsToFromMotor - battWattsToCap, pf);
			// (while it is possible to overcharge or deplete the battery, i haven't seen it happen)
			modelBatt.drawPower(battWattsToFromMotor + battWattsToCap, pf);
		} catch (PowerFlowException e) {
			// (throwing the exception here is helpful for debugging problems above)
			throw new RuntimeException(e);
		}
		return new PowerFlows(battWattsToFromMotor, capWattsToFromMotor, battWattsToCap);
	}

	public static double roundOff(double arg, double factor) {
		return Math.round(arg * factor) / factor;
	}

	/**
	 * Determines the initial direction in degrees, then returns the compass point as a String,
	 * N/NE/E/SE/S/SW/W/NW at 45 degrees apiece.
	 */
	private static final String[] COMPASS_POINTS = new String[] {
			"N", "NE", "E", "SE", "S", "SW", "W", "NW", "N"};
	private static String getCompassPoint(double latDeg1, double lonDeg1, double latDeg2, double lonDeg2) {
		double latRad1 = Math.toRadians(latDeg1);
		double latRad2 = Math.toRadians(latDeg2);
		double deltaLonRad = Math.toRadians(lonDeg2 - lonDeg1);
		// see: http://williams.best.vwh.net/avform.htm#Crs "Course between points" alternative formula:
		//   tc1=mod(atan2(sin(lon1-lon2)*cos(lat2),cos(lat1)*sin(lat2)-sin(lat1)*cos(lat2)*cos(lon1-lon2)),2*pi)
		// note: the starting point cannot be the north or south poles.
		double dirRad = Math.atan2(Math.sin(deltaLonRad) * Math.cos(latRad2),
				Math.cos(latRad1) * Math.sin(latRad2) - Math.sin(latRad1) * Math.cos(latRad2) * Math.cos(deltaLonRad));
		// direction in degrees -180.0 to +180.0
		double dirDeg = Math.toDegrees(dirRad);
		// convert to degrees 0 to 360
		dirDeg = (dirDeg + 360.0) % 360.0;
		// change direction in degrees to 8 compass points at 45 degrees each
		return COMPASS_POINTS[(int) Math.round(dirDeg / 45.0)];
	}

	public void endTrip() {
		modelCap = null;
		modelBatt = null;
	}

	public String getName() {
		return name;
	}

}
